#pragma once
/*
	Implementation of Level set numerical solver from the book 
	"Geometric Level Set Methods in Imaging, Vision and graphics", Chaper 1
	with the book
	"Level set methods and Dynamic Implicit Surface",
	It solves the level set equation
	\frac{\partial \varpi }{\partial t} + v. \theta \varpi = 0
	where 
	v = a(x) - \mu (x) div(\frac{\theta \varpi}{|\theta \varpi|})
	a(x) = a(x) * (\varpi_x, \varpi_y) / sqrt(|\varpi_x^2+\varpi_y^2+\varepsilon| is a vector

	Notations:

	x = (x,y)
	xj = j\theta x
	yk = k \theta y
	tn = n \theta t
	
	Definition for divided difference
	D_{+-}^x /varpi_jk = +- (/_j+-1,k - /_jk)  / /Delta
	D_{+-}^y /varpi_jk = +- (/_j,k+-1 - /_jk)  / /Delta

	The PDE is solved in three steps
	1) Approximate the convection term a(x).\theta \varpi
	2) Approximate the Curvature term by central difference methods	
	3) Use TVD Runge-Kutta schemes, to do the time discretization
	
	Note 1: the key is the correct and efficient implementation of the divide difference function D, 
			a generalization of which is in the book "Level set Methods and Dynamic Implicit Surface"
	Note 2: in OpenCV, the IplImage is defined as height x width (row x col)
				and the CvMat is defined as widthXheight (row x col)
	
*/

#include "LSCommon.h"

#define MINIM 1 //10e-1 //  \epsilon = 10^-10

class LSSolver
{
protected:
	bool loaded;

	int width, height;
	CvMat *ax;		// a(x)
	CvMat *mux;		// mu(x)
	CvMat *varpi;	// varpi(x)
	CvMat *varpi_x;
	CvMat *varpi_y;
	double dx,dy;	// Should be 1/gridnum ? why?
	double dt;		

	CvMat *convection_term;	 // Used to store temporary calculation result
	CvMat *curvature_term;	

	CvVideoWriter *vout;
public:	
	LSSolver::LSSolver( CvMat *varpi_i, CvMat *ax_i, CvMat *mux_i, 
		double dt_i, bool video, char* videofname);
	LSSolver::LSSolver();
	~LSSolver(void);
protected:	
	double Dx(int order, int indx, int indy, float s); // calculate the divided difference D_index^order+s /varpi of the X component
	double Dy(int order, int indx, int indy, float s); // calculate the divided difference D_index^order of the Y component
	double M(double x, double y);		
	void convectionENO2_iterate();	// order 2 ENO approximation of the convection term
	void curvature_iterate();		//curvature approximation using central difference methods
	void TVDRK2_iterate();			//order 2 TVD Runge-Kutta method
public: // Functions that can be used as global (independent of the LSSolver class)
	void ReInitDF(CvMat *disfun);		// re-initialize the distance function
	void Gradient(CvMat *I, CvMat *gradIx, CvMat *gradIy); // Calculate the gradient of the image I	
	void Extract_Contour(CvMat *disfun, CvMat *targetMat, double threshold);
	void Saturate(int &i, int &j);
	double Dx(CvMat*mat, int order, int indx, int indy, float s); // calculate the divided difference D_index^order+s /varpi of the X component
	double Dy(CvMat*mat, int order, int indx, int indy, float s); // calculate the divided difference D_index^order of the Y component
public:	// The interface
	void Iterate();						//Iterate one step of level set equation
	CvMat *GetResult();
	void output2movie(IplImage *frame);
};
