#include ".\sdistmap.h"

void defaultErrorFn(char* msg)
{
	std::cerr << msg << endl;
}

SDistMap::SDistMap(int width, int height, ErrorFunction errorFun): 
	m_height(height), m_width(width), INFINITE_D(1e10), errfunction(errorFun)
{	
	_ASSERT(width > 1 && height > 1);

	m_DMap = new double[m_width*m_height];
	m_LMap = new Label[m_width*m_height];

	m_contourFirst = m_contourLast = NULL;

	if(errfunction == NULL)
		errfunction = defaultErrorFn;

	m_curStat = TOBEGIN;	

	for(int i = 0; i<m_width*m_height; i++){
		m_DMap[i] = INFINITE_D;
		m_LMap[i] = OUTER;
	}
}

SDistMap::~SDistMap(void)
{
	delete[] m_DMap;
	delete[] m_LMap;

	contour_list_ptr pcontour;
	while(m_contourFirst != NULL) {
		pcontour = m_contourFirst;
		m_contourFirst = m_contourFirst->next;
		delete pcontour;
	}
}


/*
	Begin a new contour
*/
void SDistMap::begin_contour(int x, int y)
{ 	
	_ASSERT(x >=0 && x < m_width && y >= 0 && y < m_height);

	saturate( x, y);

	if(m_curStat == TOBEGIN)
		m_curStat = PROCEED;
	else
		return;

	if(m_contourFirst == NULL) {
		m_contourFirst = new contour_list;
		m_contourLast = m_contourFirst;
		m_contourLast->next = NULL;
	}else {
		m_contourLast->next = new contour_list;
		m_contourLast = m_contourLast->next;
		m_contourLast->next = NULL;
	}

	m_contourLast->m_prevPnt.x = x; 
	m_contourLast->m_prevPnt.y = y;
	m_contourLast->m_startPnt.x = x;
	m_contourLast->m_startPnt.y = y;
	m_contourLast->m_numRMax = -1;
	m_contourLast->m_numRMin = m_width;
	m_contourLast->m_numCMax = -1;
	m_contourLast->m_numCMin = m_height;

}

void SDistMap::add_point(int x, int y)
{
	_ASSERT(x >=0 && x < m_width && y >= 0 && y < m_height);

	saturate(x, y);
	
	if(m_curStat != PROCEED || 
		(m_contourLast->m_prevPnt.x == x && m_contourLast->m_prevPnt.y == y))
		return;	

	point prepnt = m_contourLast->m_prevPnt;

	{//////////Bresenham Algorithm Implemented here/////////////
		int tx = x;
		int ty = y;
		int signs = 1;
		if(tx == prepnt.x)
			signs = 0;
		int x0 = prepnt.x, y0 = prepnt.y;
		int dx = abs(tx-x0), dy = abs(ty-y0);
		int s1 = tri_sign(tx-x0), s2 = tri_sign(ty-y0);
		int tmp, exchange;
		if(dy > dx) {
			tmp = dx;
			dx = dy;
			dy = tmp;
			exchange = 1;
		}else {
			exchange = 0;
		}
		int e = 2*dy - dx;

		for(int i = 1; i <= dx; i++) {//(x0,y0) is not included	
			while(e > 0) {
				if(exchange == 1)
					x0 += s1;
				else
					y0 += s2;
				e = e - 2*dx;
			}
			if(exchange == 1)
				y0 += s2;
			else
				x0 += s1;
			e = e + 2*dy;

			if(x0 < m_contourLast->m_numRMin)		m_contourLast->m_numRMin = x0;
			if(x0 > m_contourLast->m_numRMax)		m_contourLast->m_numRMax = x0;
			if(y0 < m_contourLast->m_numCMin)		m_contourLast->m_numCMin = y0;
			if(y0 > m_contourLast->m_numCMax)		m_contourLast->m_numCMax = y0;

			if(getLMap(x0, y0) == OUTER) {				
				point pnt(x0, y0);						
				m_contourLast->pnts.push_back(pnt);
				setLMap(x0, y0, EDGE);
			}else{				
				errfunction("Currently no intersection of contours is supported!");
				_ASSERT(false);
			}			
		}	
	}//End of Bresenham Algorithm Implemention

	m_contourLast->m_prevPnt.x = x; 
	m_contourLast->m_prevPnt.y = y;   	
}


void SDistMap::end_contour()
{
    if(m_curStat != PROCEED)		
		return;

	add_point(m_contourLast->m_startPnt.x, m_contourLast->m_startPnt.y);

	if(m_contourLast->pnts.size() == 0) { // only one point
		point pnt(m_contourLast->m_startPnt.x, m_contourLast->m_startPnt.y);
		setLMap(pnt.x, pnt.y, EDGE);
		m_contourLast->pnts.push_back(pnt);
	}		

	m_curStat = TOBEGIN;	

	codeLMap();	
}


SDistMap::Label SDistMap::what_Label(int x, int y)
{
	_ASSERT(x >=0 && x < m_width && y >= 0 && y < m_height);

	return getLMap(x, y);
}

double SDistMap::what_Distance(int x, int y)
{
	_ASSERT(x >=0 && x < m_width && y >= 0 && y < m_height);

	return getDMap(x, y);
}

double SDistMap::what_SDistance(int x, int y)
{
	_ASSERT(x >=0 && x < m_width && y >= 0 && y < m_height);

	return getSDMap(x, y);
}

void SDistMap::code_DMap()
{
	if(m_curStat == TOBEGIN){		
		codeDMap();
	}
}

bool SDistMap::is_proceeding()
{
	return (m_curStat == PROCEED);
}
